# -*- coding: UTF-8 -*-
"""
自定义SNN的学习机制，以STDP作为样例。
"""


import torch
import torch.nn as nn
from typing import Tuple as _Tuple


@torch.jit.script
def stdp_online(delta_weight: torch.Tensor, input_trace: torch.Tensor, output_trace: torch.Tensor, input_spike_train: torch.Tensor, output_spike_train: torch.Tensor, a_pos: float, tau_pos: float, a_neg: float, tau_neg: float) -> _Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    在线脉冲时序依赖可塑性（STDP），基于所记录的迹给出更新量。
    Args:
        delta_weight (torch.Tensor): 权重矩阵，形状为[O, I]
        input_trace (torch.Tensor): 输入的迹，形状为[B, O, I]
        output_trace (torch.Tensor): 输出的迹，形状为[B, O, I]
        input_spike_train (torch.Tensor): 当前时间步的输入脉冲，形状为[B, I]
        output_spike_train (torch.Tensor): 当前时间步的输入脉冲，形状为[B, O]
        a_pos (float): STDP参数A+
        tau_pos (float): STDP参数tau+
        a_neg (float): STDP参数A-
        tau_neg (float): STDP参数tau-
    Returns:
        delta_weight (torch.Tensor): 更新权重之后的权重矩阵，形状为[O, I]
        input_trace (torch.Tensor): 输入的迹，形状为[B, O, I]
        output_trace (torch.Tensor): 输出的迹，形状为[B, O, I]
    """
    input_trace_batch_size, input_trace_output_shape, input_trace_input_shape = input_trace.shape
    output_trace_batch_size, output_trace_output_shape, output_trace_input_shape = output_trace.shape
    input_batch_size, input_shape = input_spike_train.shape
    output_batch_size, output_shape = output_spike_train.shape
    weight_output_shape, weight_input_shape = delta_weight.shape

    assert input_shape == weight_input_shape, "Unmatched input shape, %d required for weight but %d found." % (weight_input_shape, input_shape)
    assert input_shape == input_trace_input_shape, "Unmatched input shape, %d required for input trace but %d found." % (input_trace_input_shape, input_shape)
    assert input_shape == output_trace_input_shape, "Unmatched input shape, %d required for output trace but %d found." % (output_trace_input_shape, input_shape)
    assert output_shape == weight_output_shape, "Unmatched output shape, %d required but %d found." % (weight_output_shape, output_shape)
    assert output_shape == input_trace_output_shape, "Unmatched output shape, %d required for input trace but %d found." % (input_trace_output_shape, output_shape)
    assert output_shape == output_trace_output_shape, "Unmatched output shape, %d required for output trace but %d found." % (output_trace_output_shape, output_shape)
    assert output_batch_size == input_batch_size, "Unmatched batch size, %d for output but %d for input." % (output_batch_size, input_batch_size)
    assert input_trace_batch_size == input_batch_size, "Unmatched batch size, %d required for input trace but %d found." % (input_trace_batch_size, input_batch_size)
    assert output_trace_batch_size == output_batch_size, "Unmatched batch size, %d required for output trace but %d found." % (output_trace_batch_size, output_batch_size)
    batch_size = output_batch_size

    input_spike_mat = input_spike_train[:, None].repeat_interleave(output_shape, dim = 1)
    output_spike_mat = output_spike_train[:, :, None].repeat_interleave(input_shape, dim = 2)
    delta_weight += torch.sum(a_pos * output_spike_mat * input_trace - a_neg * input_spike_mat * output_trace, dim = 0)
    input_trace = 1.0 / tau_pos * input_trace + input_spike_mat
    output_trace = 1.0 / tau_pos * output_trace + output_spike_mat
    return delta_weight, input_trace, output_trace


@torch.jit.script
def stdp_py(delta_weight: torch.Tensor, input_spike_train: torch.Tensor, output_spike_train: torch.Tensor, a_pos: float, tau_pos: float, a_neg: float, tau_neg: float) -> torch.Tensor:
    """
    调用在线STDP作为STDP的python版本实现。
    Args:
        delta_weight (torch.Tensor): 权重矩阵，形状为[O, I]
        input_spike_train (torch.Tensor): 输入脉冲序列，形状为[T, B, I]
        output_spike_train (torch.Tensor): 输出脉冲序列，形状为[T, B, O]
        a_pos (float): STDP参数A+
        tau_pos (float): STDP参数tau+
        a_neg (float): STDP参数A-
        tau_neg (float): STDP参数tau-
    Returns:
        delta_weight (torch.Tensor): 权重增量
    """
    input_time_steps, input_batch_size, input_shape = input_spike_train.shape
    output_time_steps, output_batch_size, output_shape = output_spike_train.shape
    weight_output_shape, weight_input_shape = delta_weight.shape

    assert input_shape == weight_input_shape, "Unmatched input shape, %d required but %d found." % (weight_input_shape, input_shape)
    assert output_shape == weight_output_shape, "Unmatched output shape, %d required but %d found." % (weight_output_shape, output_shape)
    assert output_time_steps == input_time_steps, "Unmatched time steps, %d for output but %d for input." % (output_time_steps, input_time_steps)
    assert output_batch_size == input_batch_size, "Unmatched batch size, %d for output but %d for input." % (output_batch_size, input_batch_size)
    time_steps = output_time_steps
    batch_size = output_batch_size

    demo = delta_weight[None].repeat_interleave(batch_size, dim = 0) # [B, O, I]
    input_trace = torch.zeros_like(demo)
    output_trace = torch.zeros_like(demo)
    for t in range(time_steps):
        delta_weight, input_trace, output_trace = stdp_online(
            delta_weight = delta_weight,
            input_trace = input_trace,
            output_trace = output_trace,
            input_spike_train = input_spike_train[t],
            output_spike_train = output_spike_train[t],
            a_pos = a_pos,
            tau_pos = tau_pos,
            a_neg = a_neg,
            tau_neg = tau_neg
        )
    return delta_weight


if torch.cuda.is_available():
    try:
        from matterhorn_cuda_extensions import cu_stdp as stdp_cuda
    except:
        stdp_cuda = None
try:
    from matterhorn_cpp_extensions import stdp as stdp_cpp
except:
    stdp_cpp = None


def stdp(delta_weight: torch.Tensor, input_spike_train: torch.Tensor, output_spike_train: torch.Tensor, a_pos: float, tau_pos: float, a_neg: float, tau_neg: float, precision: float = 1e-6) -> torch.Tensor:
    """
    脉冲时序依赖可塑性（STDP）函数。
    Args:
        delta_weight (torch.Tensor): 权重矩阵，形状为[O, I]
        input_spike_train (torch.Tensor): 输入脉冲序列，形状为[T, B, I]
        output_spike_train (torch.Tensor): 输出脉冲序列，形状为[T, B, O]
        a_pos (float): STDP参数A+
        tau_pos (float): STDP参数tau+
        a_neg (float): STDP参数A-
        tau_neg (float): STDP参数tau-
        precision (float): 精度，将小于这一数值的值（大多由误差导致）转化为0
    Returns:
        delta_weight: 权重更新值
    """
    input_time_steps, input_batch_size, input_shape = input_spike_train.shape
    output_time_steps, output_batch_size, output_shape = output_spike_train.shape
    weight_output_shape, weight_input_shape = delta_weight.shape

    assert input_shape == weight_input_shape, "Unmatched input shape, %d required but %d found." % (weight_input_shape, input_shape)
    assert output_shape == weight_output_shape, "Unmatched output shape, %d required but %d found." % (weight_output_shape, output_shape)
    assert output_time_steps == input_time_steps, "Unmatched time steps, %d for output but %d for input." % (output_time_steps, input_time_steps)
    assert output_batch_size == input_batch_size, "Unmatched batch size, %d for output but %d for input." % (output_batch_size, input_batch_size)
    time_steps = output_time_steps
    batch_size = output_batch_size

    w_dev = delta_weight.device
    i_dev = input_spike_train.device
    o_dev = output_spike_train.device
    assert w_dev == i_dev, "Unmatched device, %s for weight but %s for input." % (w_dev, i_dev)
    assert w_dev == o_dev, "Unmatched device, %s for weight but %s for output." % (w_dev, o_dev)
    device_type = w_dev.type
    device_idx = w_dev.index

    delta_weight = torch.zeros_like(delta_weight)
    if device_type == "cuda" and stdp_cuda is not None:
        stdp_cuda(delta_weight, input_shape, output_shape, time_steps, input_spike_train, output_spike_train, a_pos, tau_pos, a_neg, tau_neg, batch_size)
    # elif stdp_cpp is not None:
    #     delta_weight_cpu = delta_weight.cpu()
    #     stdp_cpp(delta_weight_cpu, input_shape, output_shape, time_steps, input_spike_train.cpu(), output_spike_train.cpu(), a_pos, tau_pos, a_neg, tau_neg, batch_size)
    #     delta_weight = delta_weight_cpu.to(delta_weight)
    else:
        delta_weight = stdp_py(delta_weight, input_spike_train, output_spike_train, a_pos, tau_pos, a_neg, tau_neg)
    delta_weight = torch.where(torch.abs(delta_weight) >= precision, delta_weight, torch.zeros_like(delta_weight))
    return delta_weight


# -*- coding: UTF-8 -*-
"""
脉冲神经网络的解码机制。
注意：此单元可能会改变张量形状。
"""


import torch
import torch.nn as nn
from matterhorn_pytorch.__func__ import transpose as _transpose
from matterhorn_pytorch.snn.skeleton import Module as _Module
from typing import Callable as _Callable


class Decoder(_Module):
    def __init__(self) -> None:
        """
        解码器的基类。解码器是一个多时间步模型。
        """
        super().__init__()
        self.multi_step_mode_()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ""


    @property
    def supports_single_step_mode(self) -> bool:
        """
        是否支持单个时间步。
        Returns:
            if_support (bool): 是否支持单个时间步
        """
        return False


class SumSpike(Decoder):
    def __init__(self) -> None:
        """
        取张量在时间维度上的总值（总脉冲）。
        $$o_{i}=\sum_{t=1}^{T}{O_{i}^{K}(t)}$$
        """
        super().__init__()
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[T,B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        y = x.sum(dim = 0)
        return y


class AverageSpike(Decoder):
    def __init__(self) -> None:
        """
        取张量在时间维度上的平均值（平均脉冲）。
        $$o_{i}=\frac{1}{T}\sum_{t=1}^{T}{O_{i}^{K}(t)}$$
        """
        super().__init__()
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[T,B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        y = x.mean(dim = 0)
        return y


class TimeBased(Decoder):
    def __init__(self, empty_fill: float = -1, transform: _Callable = lambda x: x) -> None:
        """
        基于时间的解码器。
        Args:
            empty_fill (float): 如果脉冲序列为全0序列，值应该用什么替代，在TNN中该参数应设为torch.inf
            transform (Callable): 将结果y如何变形
        """
        super().__init__()
        self.empty_fill = empty_fill
        self.transform = transform
        self.reset()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["empty_fill=%g" % self.empty_fill])


    def reset(self) -> _Module:
        """
        重置编码器。
        """
        self.current_time_step = 0
        return super().reset()


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[T,B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        y = self.f_decode(x).to(x)
        y = y.to(x)
        mask = x.mean(dim = 0) > 0
        y = torch.where(mask, y, torch.full_like(y, self.empty_fill))
        self.current_time_step += x.shape[0]
        y = self.transform(y)
        return y


class MinTime(TimeBased):
    def __init__(self, empty_fill: float = -1, transform: _Callable = lambda x: x) -> None:
        """
        取张量在时间维度上的最小值。
        $$o_{i}=min(tO_{i}^{K}(t))$$
        Args:
            empty_fill (float): 如果脉冲序列为全0序列，值应该用什么替代，在TNN中该参数应设为torch.inf
            transform (Callable): 将结果y如何变形
        """
        super().__init__(
            empty_fill = empty_fill,
            transform = transform
        )


    def f_decode(self, x: torch.Tensor) -> torch.Tensor:
        """
        解码函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[T,B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        y = torch.argmax(x, dim = 0) + self.current_time_step
        return y


class AverageTime(TimeBased):
    def __init__(self, empty_fill: float = -1, transform: _Callable = lambda x: x) -> None:
        """
        取张量在时间维度上的时间加权平均值
        $$o_{i}=\frac{1}{T}\sum_{t=1}^{T}{tO_{i}^{K}(t)}$$
        Args:
            empty_fill (float): 如果脉冲序列为全0序列，值应该用什么替代，在TNN中该参数应设为torch.inf
            transform (Callable): 将结果y如何变形
        """
        super().__init__(
            empty_fill = empty_fill,
            transform = transform
        )


    def f_decode(self, x: torch.Tensor) -> torch.Tensor:
        """
        解码函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[T,B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        xt = _transpose(_transpose(x) * torch.arange(x.shape[0]).to(x))
        tsum = torch.sum(xt, dim = 0)
        xsum = torch.sum(x, dim = 0)
        mask = xsum > 0
        y = torch.where(mask, tsum / xsum, torch.zeros_like(xsum))
        return y


# -*- coding: UTF-8 -*-
"""
此文件夹放置SNN的一些基本模块。
"""


from . import functional
from .container import Spatial, Temporal, Sequential, ModuleList, ModuleDict, Agent
from .decoder import SumSpike as SumSpikeDecoder, AverageSpike as AvgSpikeDecoder, MinTime as MinTimeDecoder, AverageTime as AvgTimeDecoder
from .encoder import Direct as DirectEncoder, Poisson as PoissonEncoder, Temporal as TemporalEncoder
from .layer import STDPLinear, STDPConv2d, MaxPool1d, MaxPool2d, MaxPool3d, AvgPool1d, AvgPool2d, AvgPool3d, MaxUnpool1d, MaxUnpool2d, MaxUnpool3d, Upsample, Flatten, Unflatten, Dropout, Dropout1d, Dropout2d, Dropout3d
from .skeleton import Module
from .soma import IF, LIF, QIF, ExpIF, Izhikevich, KLIF, LIAF
from .surrogate import Rectangular, Polynomial, Sigmoid, Gaussian
from .synapse import Linear, Conv1d, Conv2d, Conv3d, ConvTranspose1d, ConvTranspose2d, ConvTranspose3d, BatchNorm1d, BatchNorm2d, BatchNorm3d, LayerNorm, Identity, Neurotransmitter


# -*- coding: UTF-8 -*-
"""
此文件夹放置各式各样的小工具。
"""


from matterhorn_pytorch.__func__ import cpp_available, cuda_available, transpose
from . import lsm
from . import snn
from . import tnn
from . import training
from . import util


from setuptools import find_packages
from setuptools import setup
import os
# import platform
# from typing import List


# def get_cpp_files(root_path: str, exceptions: List[str]) -> List[str]:
#     full_list = os.listdir(root_path)
#     res_list = []
#     for filename in full_list:
#         if (filename.endswith(".cpp") or filename.endswith(".cu")) and not (filename in exceptions):
#             res_list.append(os.path.join(root_path, filename))
#     return res_list


requirements = ["torch"]
# try:
#     ext_modules = []
#     from torch.cuda import is_available
#     cuda_available = is_available()
#     if cuda_available:
#         print("\033[92mCUDA found on this device, installing Matterhorn CUDA extensions.\033[0m")
#         from torch.utils.cpp_extension import CUDAExtension
#         files = get_cpp_files(os.path.join(os.path.abspath("."), "matterhorn_cuda_extensions"), ["base.cpp", "base.cu"])
#         print("Will compile " + ", ".join(files))
#         ext_modules.append(CUDAExtension(
#             "matterhorn_cuda_extensions",
#             files,
#             extra_compile_args = {
#                 "cxx": ["-g", "-w"],
#                 "nvcc": ["-O2"]
#             }
#         ))
#     else:
#         print("\033[93mCUDA not found on this device. If you have NVIDIA's GPU, please install CUDA and try again later, or manually install Matterhorn CUDA extensions.\033[0m")
#     cpp_available = not os.system("g++ --version")
#     if cpp_available:
#         print("\033[92mG++ found on this device, installing Matterhorn CPP extensions.\033[0m")
#         from torch.utils.cpp_extension import CppExtension
#         files = get_cpp_files(os.path.join(os.path.abspath("."), "matterhorn_cpp_extensions"), ["base.cpp"])
#         print("Will compile " + ", ".join(files))
#         ext_modules.append(CppExtension(
#             "matterhorn_cpp_extensions",
#             files,
#             extra_compile_args = {
#                 "cxx": ["-g", "-w"]
#             }
#         ))
#     else:
#         print("\033[93mG++ not found on this device. Please install G++ and try again later, or manually install Matterhorn CPP extensions.\033[0m")
#         if platform.system() == "Windows":
#             print("\033[93mIf you're Windows user, please download and install G++ from https://github.com/xjtuiair-cag/Matterhorn/blob/main/mingw-get-setup.exe .\033[0m")
            
#     cmdclass = {}
#     if cuda_available or cpp_available:
#         print("\033[92mTrying to build Matterhorn extensions.\033[0m")
#         if platform.system() == "Windows":
#             print("\033[93mIf you're Windows user and errors occur, please follow Microsoft's instruction and install VS generation tools for C++, then retry.\033[0m")
#         from torch.utils.cpp_extension import BuildExtension
#         cmdclass = {
#             "build_ext": BuildExtension
#         }
# except:
#     print("\033[93mFailed to build Matterhorn extensions. You can manually build Matterhorn extensions later.\033[0m")
#     ext_modules = []
#     cmdclass = {}


with open(os.path.join(os.path.abspath("."), "requirements.txt"), "r", encoding="utf-8") as fh:
    install_requires = fh.read()


with open(os.path.join(os.path.abspath("."), "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    install_requires = install_requires,
    name = "matterhorn_pytorch",
    version = "1.3.1",
    author = "CAG, IAIR, XJTU, Xi'an, China",
    author_email = "ericwang017@stu.xjtu.edu.cn",
    description = "Matterhorn is a novel general neuromorphic computing framework based on PyTorch.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    license = "License :: OSI Approved :: MIT License",
    url = "https://github.com/xjtuiair-cag/Matterhorn",
    packages = find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.7',
    # ext_modules = ext_modules,
    # cmdclass = cmdclass
)


# -*- coding: UTF-8 -*-
"""
液体状态机。
使用邻接矩阵表示神经元的连接方向。
"""


import torch
import torch.nn as nn
import torch.nn.functional as _F
from typing import Any as _Any, Tuple as _Tuple, Callable as _Callable
import matterhorn_pytorch.snn.functional as _SF
from matterhorn_pytorch.snn.skeleton import Module as _Module
from matterhorn_pytorch.snn.soma import Soma as _Soma
from matterhorn_pytorch.snn.container import Temporal as _Temporal
from matterhorn_pytorch.training.functional import stdp_online as _stdp_online


class LSM(_Module):
    def __init__(self, adjacent: torch.Tensor, soma: _Soma, device: torch.device = None, dtype: torch.dtype = None) -> None:
        """
        液体状态机。
        Args:
            adjacent (torch.Tensor): 邻接矩阵，行为各个神经元的轴突，列为各个神经元的树突，1为有从轴突指向树突的连接，0为没有
            soma (nn.Module): 使用的脉冲神经元胞体，在matterhorn_pytorch.snn.soma中选择
            device (torch.device): 所计算的设备
            dtype (torch.dtype): 所计算的数据类型
        """
        assert len(adjacent.shape) == 2 and adjacent.shape[0] == adjacent.shape[1], "Incorrect adjacent matrix."
        super().__init__()
        self.o = None
        self.adjacent = nn.Parameter(adjacent.to(torch.float), requires_grad = False)
        self.soma = soma
        self.neuron_num = self.adjacent.shape[0]
        self.weight = nn.Parameter(torch.empty((self.neuron_num, self.neuron_num), device = device, dtype = dtype))
        nn.init.kaiming_uniform_(self.weight, a = 5.0 ** 0.5)


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["neuron_num=%d" % self.neuron_num, "adjacent=\n%s\n" % self.adjacent]) + ((", " + super().extra_repr()) if len(super().extra_repr()) else "")


    def reset(self) -> _Module:
        """
        重置整个神经元。
        """
        if isinstance(self.soma, _Module):
            self.soma.reset()
        self.o = _SF.reset_tensor(self.o, 0.0)
        return super().reset()


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        单个时间步的前向传播函数。
        Args:
            x (torch.Tensor): 来自突触的输入电位$X_{i}^{l}(t)$
        Returns:
            o (torch.Tensor): 胞体当前的输出脉冲$O_{i}^{l}(t)$
        """
        self.o = _SF.init_tensor(self.o, x)
        y = x + _F.linear(self.o, self.weight * self.adjacent.T, None)
        o = self.soma.forward_single_time_step(y)
        self.o = o
        return o


# -*- coding: UTF-8 -*-
"""
脉冲神经网络的编码机制。
注意：此单元可能会改变张量形状。
"""


import torch
import torch.nn as nn
import matterhorn_pytorch.snn.functional as _SF
from matterhorn_pytorch.snn.skeleton import Module as _Module
from typing import Callable as _Callable


class Encoder(_Module):
    def __init__(self) -> None:
        """
        编码器的基类。编码器是一个多时间步模型。
        """
        super().__init__()
        self.multi_step_mode_()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ""


    @property
    def supports_single_step_mode(self) -> bool:
        """
        是否支持单个时间步。
        Returns:
            if_support (bool): 是否支持单个时间步
        """
        return False


class Direct(Encoder):
    def __init__(self) -> None:
        """
        直接编码，直接对传入的脉冲（事件）数据进行编码
        """
        super().__init__()
    

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        直接编码的前向传播函数，直接将第二个维度作为时间维度，转置到首个维度上
        Args:
            x (torch.Tensor): 输入张量，形状为[B,T,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[T,B,...]
        """
        idx = [i for i, j in enumerate(x.shape)]
        assert len(idx) >= 2, "There is no temporal dimension."
        idx[0], idx[1] = idx[1], idx[0]
        y = x.permute(*idx)
        return y


class Poisson(Encoder):
    def __init__(self, time_steps: int = 1) -> None:
        """
        泊松编码（速率编码），将值转化为脉冲发放率（多步）
        Args:
            time_steps (int): 生成的时间步长
        """
        super().__init__()
        self.time_steps = time_steps


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["time_steps=%d" % self.time_steps])


    def forward_single_time_step(self, x:torch.Tensor) -> torch.Tensor:
        """
        单步前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        r = torch.rand_like(x)
        y = _SF.le(r, x)
        return y
    

    def forward_multi_time_steps(self, x: torch.Tensor) -> torch.Tensor:
        """
        多步前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[T, B,...]
        """
        res_shape = [self.time_steps] + list(x.shape)
        v = torch.ones(*res_shape).to(x) * x
        y = self.forward_single_time_step(v)
        return y


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        泊松编码的前向传播函数，将值$V$转化为该时间步$t$内的脉冲$O^{0}(t)$
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[T,B,...]
        """
        min_value = torch.min(x)
        max_value = torch.max(x)
        if max_value == min_value:
            x = torch.full_like(x, 0.5)
        else:
            x = (x - min_value) / (max_value - min_value)
        if self.time_steps <= 1:
            y = self.forward_single_time_step(x)
        else:
            y = self.forward_multi_time_steps(x)
        return y


class Temporal(Encoder):
    def __init__(self, time_steps: int = 1, prob: float = 1.0, transform: _Callable = lambda x: x) -> None:
        """
        时间编码，在某个时间之前不会产生脉冲，在某个时间之后随机产生脉冲
        Args:
            time_steps (int): 生成的时间步长
            prob (float): 若达到了时间，以多大的概率发放脉冲，范围为[0, 1]
            transform (Callable): 将数据x如何变形
        """
        super().__init__()
        self.time_steps = time_steps
        self.prob = prob
        self.transform = transform
        self.reset()


    def extra_repr(self) -> str:
        """
        额外的表达式，把参数之类的放进来。
        Returns:
            repr_str (str): 参数表
        """
        return ", ".join(["time_steps=%d" % self.time_steps, "prob=%g" % self.prob])


    def reset(self) -> _Module:
        """
        重置编码器
        """
        self.current_time_step = 0
        return super().reset()


    def forward_single_time_step(self, x: torch.Tensor) -> torch.Tensor:
        """
        单步前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[B,...]
        """
        f = _SF.le(x, self.current_time_step)
        r = _SF.le(torch.rand_like(x), self.prob)
        y = f * r
        self.current_time_step += 1
        return y
    

    def forward_multi_time_steps(self, x: torch.Tensor, time_steps: int) -> torch.Tensor:
        """
        多步前向传播函数。
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[T, B,...]
        """
        y_seq = []
        for t in range(time_steps):
            y_seq.append(self.forward_single_time_step(x))
        y = torch.stack(y_seq)
        return y


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        时间编码的前向传播函数，将值$V$转化为该时间步$t$内的脉冲$O^{0}(t)$
        Args:
            x (torch.Tensor): 输入张量，形状为[B,...]
        Returns:
            y (torch.Tensor): 输出张量，形状为[T,B,...]
        """
        x = self.transform(x)
        if self.time_steps <= 1:
            y = self.forward_single_time_step(x)
        else:
            y = self.forward_multi_time_steps(x, self.time_steps)
        return y